import minako.node.*;
import minako.analysis.*;
import java.util.*;

public class Semantic extends DepthFirstAdapter {
	
	private int maincount = 0;
	private SymbolTable symtab = new SymbolTable();
	private Codegen codegen = new Codegen();
	private Vector params = new Vector();
	private String paramString = "";
	private String paramStringCall = "";
	private Stack labelStack = new Stack();
	private Stack codeAdressStack = new Stack();

	String convertType (int type) {
		String typestring = "";
		if (type == Minako.TYP_BOOLEAN) {
			typestring = "Z";
		} else if (type == Minako.TYP_INT) {
			typestring = "I";
		} else if (type == Minako.TYP_FLOAT) {
			typestring = "F";
		} else {
			typestring = "" + type;
		}
		return typestring;
	}
	
	String convertType (String type) {
		String typestring = "";
		if (type.equals("3")) {
			typestring = "Z";
		} else if (type.equals("1")) {
			typestring = "I";
		} else if (type.equals("2")) {
			typestring = "F";
		} else {
			typestring = "" + type;
		}
		return typestring;
	}
	
	void makeAssignment(SymbolTableEntry entry) {
		StringBuffer tempString = new StringBuffer();
		if (entry.getLevel() != 1) {
			if ((entry.getType() == Minako.TYP_INT) || (entry.getType() == Minako.TYP_BOOLEAN)) tempString.append("istore " + entry.getAddress());
			else if (entry.getType() == Minako.TYP_FLOAT) tempString.append("fstore " + entry.getAddress());
			else Minako.error("compiler-intern error in makeAssignment.");
			codegen.appendCommand(tempString.toString(), -1);
		}
		else {
			if (entry.getType() == Minako.TYP_INT) tempString.append("putstatic " + Minako.className + "/" + entry.getName() + " I");
			else if (entry.getType() == Minako.TYP_FLOAT) tempString.append("putstatic " + Minako.className + "/" + entry.getName() + " F");
			else if (entry.getType() == Minako.TYP_BOOLEAN) tempString.append("putstatic " + Minako.className + "/" + entry.getName() + " Z");
			else Minako.error("compiler-intern error in makeAssignment.");
			codegen.appendCommand(tempString.toString(), -1);
		}
	}

	void matchTypes(int type1, int type2) {
		if ((type1 != type2) && (!((type1 == Minako.TYP_FLOAT) && (type2 == Minako.TYP_INT)))) {
			Minako.error("Type mismatch.\n");
		}
	}

	public int combineTypes(int type1, int type2, String op) {
		
		/* no operation allows void-arguments */
		if ((type1 == Minako.TYP_VOID) || (type2 == Minako.TYP_VOID)) {
			Minako.error("Void not allowed in arithmetical/logical operation.\n");
		}
		/* eq and neq allow integer, floats and booleans; geq, leq, ... only integer and floatsnly integer and floats */
		else if ((op == "==") || (op == "!=") || (op == "<=") || (op == ">=") || (op == "<") || (op == ">")) {

			if ( ((op == "<=") || (op == ">=") || (op == "<") || (op == ">"))
									&& ((type1 == Minako.TYP_BOOLEAN) || (type2 == Minako.TYP_BOOLEAN)) ) {
				Minako.error("Boolean not allowed in arithmetical comparison.\n");
			}
			if ((type1 == Minako.TYP_BOOLEAN) || (type2 == Minako.TYP_BOOLEAN)) {
				if ((type1 != Minako.TYP_BOOLEAN) || (type2 != Minako.TYP_BOOLEAN)) {
					Minako.error("Booleans can only be compared with other Booleans.\n");
				}
			}
			return Minako.TYP_BOOLEAN;
		}
		/* or and and are boolean-only operations */
		else if ((op == "||") || (op == "&&")) {
			if ((type1 != Minako.TYP_BOOLEAN) || (type2 != Minako.TYP_BOOLEAN)) Minako.error("Boolean expected for logical operation.\n");
			return Minako.TYP_BOOLEAN;
		}
		/* operations for numbers... */
		else if ((op == "+") || (op == "-") || (op == "*") || (op == "/")) {
			if ((type1 == Minako.TYP_BOOLEAN) || (type2 == Minako.TYP_BOOLEAN)) {
				Minako.error("Boolean not allowed in arithmetical operation.\n");
			}
			else if ((type1 == Minako.TYP_FLOAT) || (type2 == Minako.TYP_FLOAT)) {
				return Minako.TYP_FLOAT;
			}
			else {
				return Minako.TYP_INT;
			}
		}
		/* just to make sure we haven't forget something ;-) */
		else {
			Minako.error("Unknown operation (compiler-intern error)\n");
		}
		/* just to avoid a warning */
		return 0;
	}
	
	public int getTypeType(PType type) {
		if (type instanceof AIntType) {
			return Minako.TYP_INT;
		}
		else if (type instanceof AFloatType) {
			return Minako.TYP_FLOAT;
		}
		else if (type instanceof ABooleanType) {
			return Minako.TYP_BOOLEAN;
		}
		else if (type instanceof AVoidType) {
			return Minako.TYP_VOID;
		}
		else {
			return -1;
		}
	}
	
	public int getTypeFunctioncall(PFunctioncall functioncall) {
		if (functioncall instanceof AEmptyFunctioncall) {
			return getTypeFunctioncall((AEmptyFunctioncall) functioncall);
		}
		else if (functioncall instanceof ANotEmptyFunctioncall) {
			return getTypeFunctioncall((ANotEmptyFunctioncall) functioncall);
		}
		else {
			return -1;
		}
	}
	
	public int getTypeFunctioncall(AEmptyFunctioncall functioncall) {
		return getTypeId(functioncall.getId());
	}
	
	public int getTypeFunctioncall(ANotEmptyFunctioncall functioncall) {
		return getTypeId(functioncall.getId());
	}
	
	public int getTypeAssignment(PAssignment assignment) {
		if (assignment instanceof AAssAssignment) {
			return getTypeAssignment((AAssAssignment) assignment);
		}
		else if (assignment instanceof AExprAssignment) {
			return getTypeAssignment((AExprAssignment) assignment);
		}
		else {
			return -1;
		}
	}
	
	public int getTypeAssignment(AAssAssignment assignment) {
		if (getTypeId(assignment.getId()) == Minako.TYP_FLOAT && getTypeAssignment(assignment.getAssignment()) == Minako.TYP_INT) {
			return Minako.TYP_FLOAT;
		} else {
			return getTypeAssignment(assignment.getAssignment());
		}
	}
	
	public int getTypeAssignment(AExprAssignment assignment) {
		return getTypeExpr(assignment.getExpr());
	}
	
	public int getTypeExpr(PExpr expr) {
		if (expr instanceof ATermExpr) {
			return getTypeExpr((ATermExpr) expr);
		}
		else if (expr instanceof AMinTermExpr) {
			return getTypeExpr((AMinTermExpr) expr);
		}
		else if (expr instanceof AEqExpr) {
			return getTypeExpr((AEqExpr) expr);
		}
		else if (expr instanceof ANeqExpr) {
			return getTypeExpr((ANeqExpr) expr);
		}
		else if (expr instanceof ALssExpr) {
			return getTypeExpr((ALssExpr) expr);
		}
		else if (expr instanceof AGrtExpr) {
			return getTypeExpr((AGrtExpr) expr);
		}
		else if (expr instanceof ALeqExpr) {
			return getTypeExpr((ALeqExpr) expr);
		}
		else if (expr instanceof AGeqExpr) {
			return getTypeExpr((AGeqExpr) expr);
		}
		else if (expr instanceof APlusExpr) {
			return getTypeExpr((APlusExpr) expr);
		}
		else if (expr instanceof AMinusExpr) {
			return getTypeExpr((AMinusExpr) expr);
		}
		else if (expr instanceof AOrExpr) {
			return getTypeExpr((AOrExpr) expr);
		}
		else {
			return -1;
		}
	}
	
	public int getTypeExpr(ATermExpr expr) {
		return getTypeTerm(expr.getTerm());
	}
	
	public int getTypeExpr(AMinTermExpr expr) {
		if ((getTypeTerm(expr.getTerm()) != Minako.TYP_INT) || (getTypeTerm(expr.getTerm()) != Minako.TYP_FLOAT))
			Minako.error("Prefix minus just allowed for numbers");
		return getTypeTerm(expr.getTerm());
	}
	
	public int getTypeExpr(AEqExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getEq().getText());
	}
	
	public int getTypeExpr(ANeqExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getNeq().getText());
	}
	
	public int getTypeExpr(ALssExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getLss().getText());
	}
	
	public int getTypeExpr(AGrtExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getGrt().getText());
	}
	
	public int getTypeExpr(ALeqExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getLeq().getText());
	}
	
	public int getTypeExpr(AGeqExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getGeq().getText());
	}
	
	public int getTypeExpr(APlusExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getPlus().getText());
	}
	
	public int getTypeExpr(AMinusExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getMinus().getText());
	}
	
	public int getTypeExpr(AOrExpr expr) {
		int left = getTypeExpr(expr.getExpr());
	    int right = getTypeTerm(expr.getTerm());
	       
	    return combineTypes(left, right, expr.getOr().getText());
	}
	
	public int getTypeTerm(PTerm term) {
		if (term instanceof AFactorTerm) {
			return getTypeTerm((AFactorTerm) term);
		}
		else if (term instanceof AMulTerm) {
			return getTypeTerm((AMulTerm) term);
		}
		else if (term instanceof ADivTerm) {
			return getTypeTerm((ADivTerm) term);
		}
		else if (term instanceof AAndTerm) {
			return getTypeTerm((AAndTerm) term);
		}
		else {
			return -1;
		}
	}
	
	public int getTypeTerm(AFactorTerm term) {
		return getTypeFactor(term.getFactor());
	}
	
	public int getTypeTerm(AMulTerm term) {
		int left = getTypeTerm(term.getTerm());
	    int right = getTypeFactor(term.getFactor());
	       
	    return combineTypes(left, right, term.getAsteriks().getText());
	}
	
	public int getTypeTerm(ADivTerm term) {
		int left = getTypeTerm(term.getTerm());
	    int right = getTypeFactor(term.getFactor());
	       
	    return combineTypes(left, right, term.getSlash().getText());
	}
	
	public int getTypeTerm(AAndTerm term) {
		int left = getTypeTerm(term.getTerm());
	    int right = getTypeFactor(term.getFactor());
	       
	    return combineTypes(left, right, term.getAnd().getText());
	}
	
	public int getTypeFactor(PFactor fac) {
		if (fac instanceof AIntFactor) {
			return getTypeFactor((AIntFactor) fac);
		}
		else if (fac instanceof AFloatFactor) {
			return getTypeFactor((AFloatFactor) fac);
		}
		else if (fac instanceof ABooleanFactor) {
			return getTypeFactor((ABooleanFactor) fac);
		}
		else if (fac instanceof AAssFactor) {
			return getTypeFactor((AAssFactor) fac);
		}
		else if (fac instanceof AIdFactor) {
			return getTypeFactor((AIdFactor) fac);
		}
		else if (fac instanceof AFunccallFactor) {
			return getTypeFactor((AFunccallFactor) fac);
		}
		else {
			return -1;
		}
	}
	
	public int getTypeFactor(AAssFactor fac) {
		return getTypeAssignment(fac.getAssignment());
	}
	
	public int getTypeFactor(AIdFactor fac) {
		return getTypeId(fac.getId());
	}
	
	public int getTypeFactor(AIntFactor fac) {
		return Minako.TYP_INT;
	}
	
	public int getTypeFactor(AFloatFactor fac) {
		return Minako.TYP_FLOAT;
	}
	
	public int getTypeFactor(ABooleanFactor fac) {
		return Minako.TYP_BOOLEAN;
	}
	
	public int getTypeFactor(AFunccallFactor fac) {
		return getTypeFunctioncall(fac.getFunctioncall());
	}
	
	public int getTypeId(PId id) {
		return getTypeId((AId) id);
	}
	
	public int getTypeId(AId id) {
			String temp = id.getIdent().getText();
			return symtab.getSymbolTableEntry(temp).getType();
	}
	
	public void inAProgram (AProgram node) {
		symtab.enterBlock();
		codegen.init(Minako.inputfile);
	}
	
	public void outAProgram (AProgram node) {
		symtab.leaveBlock();
		codegen.finishCode();
	}
	
	public void outStart (Start node) {
		if (maincount != 1)
			Minako.error("No main method");
	}
	
	public void inACompoundStatement (ACompoundStatement node) {
		symtab.enterBlock();
	}
	
	public void outACompoundStatement (ACompoundStatement node) {
		symtab.leaveBlock();
	}
	
	public void outAAssAssignment (AAssAssignment node) {
		matchTypes(getTypeId(node.getId()),getTypeAssignment(node.getAssignment()));
		AId id = (AId) node.getId();
		SymbolTableEntry temp = symtab.getSymbolTableEntry(id.getIdent().getText());
		if (temp.getIdClass() == SymbolTable.CLASS_FUNC)
			Minako.error("Function cannot get a value");
		codegen.appendCommand("dup",1);
		makeAssignment(symtab.getSymbolTableEntry(id.getIdent().getText()));
	}
	
	public void outAStatassignment (AStatassignment node) {
		matchTypes(getTypeId(node.getId()),getTypeAssignment(node.getAssignment()));
		AId id = (AId) node.getId();
		SymbolTableEntry temp = symtab.getSymbolTableEntry(id.getIdent().getText());
		if (temp.getIdClass() == SymbolTable.CLASS_FUNC)
			Minako.error("Function cannot get a value");
		makeAssignment(symtab.getSymbolTableEntry(id.getIdent().getText()));
	}
	
	public void outANoassignDeclassignment(ANoassignDeclassignment node) {
		if (getTypeType(node.getType()) == Minako.TYP_VOID)
			Minako.error("No assignment to void-type");
		AId id = (AId) node.getId();
		symtab.put(id.getIdent().getText(),getTypeType(node.getType()),SymbolTable.CLASS_VAR);
		codegen.addVariable(id.getIdent().getText(),getTypeType(node.getType()));
	}
	
	public void outAAssignDeclassignment(AAssignDeclassignment node) {
		matchTypes(getTypeType(node.getType()),getTypeAssignment(node.getAssignment()));
		AId id = (AId) node.getId();
		symtab.put(id.getIdent().getText(),getTypeType(node.getType()),SymbolTable.CLASS_VAR);
		codegen.addVariable(id.getIdent().getText(),getTypeType(node.getType()));
		makeAssignment(symtab.getSymbolTableEntry(id.getIdent().getText()));
		
	}
	
	public void inAPrintf(APrintf node) {
		codegen.appendCommand("getstatic java/lang/System/out Ljava/io/PrintStream;",1);
	}
	
	public void outAPrintf(APrintf node) {
		if (getTypeAssignment(node.getAssignment()) == Minako.TYP_VOID)
			Minako.error("Void-Type not printable");
		codegen.appendCommand("invokevirtual java/io/PrintStream/println(" + convertType(getTypeAssignment(node.getAssignment())) + ")V",-1);
	}
	
	public void outAReturnstatement(AReturnstatement node) {
		int type = getTypeAssignment(node.getAssignment());
		if (node.getAssignment() != null) {
			matchTypes(symtab.getSymbolTableEntryOfCurrentFunction().getType(), type);
		}
		if (type == Minako.TYP_INT || type == Minako.TYP_BOOLEAN) {
			codegen.appendCommand("ireturn",0);
		} else if (type == Minako.TYP_FLOAT) {
			codegen.appendCommand("freturn",0);
		}
	}
	
	public void caseAFunctiondefinition(AFunctiondefinition node)
    {
        inAFunctiondefinition(node);
        if(node.getType() != null)
        {
            node.getType().apply(this);
        }
        if(node.getId() != null)
        {
            node.getId().apply(this);
        }
        AId id = (AId) node.getId();
		symtab.put(id.getIdent().getText(),getTypeType(node.getType()),SymbolTable.CLASS_FUNC);
		symtab.enterBlock();
		if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        SymbolTableEntry current = symtab.getSymbolTableEntryOfCurrentFunction();
        paramString = "";
        if (current.getName().equals("main"))
			maincount = 1;
        if (current.getName().equals("main") && (current.getType() != Minako.TYP_VOID))
    		Minako.error("main has to be void");
        if(node.getParameterlist() != null)
        {
            node.getParameterlist().apply(this);
        }
        codegen.enterFunction(current.getName(),current.getType(),paramString);
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        if(node.getCompoundStatement() != null)
        {
            node.getCompoundStatement().apply(this);
        }
        symtab.leaveBlock();
        codegen.leaveFunction();
        outAFunctiondefinition(node);
    }
	
	public void outAParameterlist(AParameterlist node) {
		StringBuffer tempString = new StringBuffer();
		AId id = (AId) node.getId();
		if (getTypeType(node.getType()) == Minako.TYP_VOID)
			Minako.error("params may not be void");
		SymbolTableEntry current = symtab.getSymbolTableEntryOfCurrentFunction();
		if (current.getName().equals("main"))
			Minako.error("Main may not have params");
		current.setParamList2(0, symtab.put(id.getIdent().getText(),getTypeType(node.getType()),SymbolTable.CLASS_PARAM));
		current.incParamIndex();
		for (int i = 0;i < current.getParamIndex();i++) {
			tempString.append(convertType(current.getParamList2(i).getType()));
		}
		paramString = tempString.toString();
	}
	
	public void outAParamlisttemp(AParamlisttemp node) {
		AId id = (AId) node.getId();
		if (getTypeType(node.getType()) == Minako.TYP_VOID)
			Minako.error("params may not be void");
		SymbolTableEntry current = symtab.getSymbolTableEntryOfCurrentFunction();
		current.incParamIndex();
		current.setParamList2(current.getParamIndex(), symtab.put(id.getIdent().getText(),getTypeType(node.getType()),SymbolTable.CLASS_PARAM));
	}
	
	public void outAEmptyFunctioncall(AEmptyFunctioncall node) {
		AId id = (AId) node.getId();
		StringBuffer tempString = new StringBuffer();
		String name = id.getIdent().getText();
		SymbolTableEntry temp = symtab.getSymbolTableEntry(name);
		if (temp.getParamIndex() != 0)
			Minako.error("Function has params");
		tempString.append("invokestatic " + Minako.className + "/" + temp.getName() + "()" + convertType(temp.getType()));
		codegen.appendCommand(tempString.toString(),1);
		if (temp.getType() != Minako.TYP_VOID && node.parent() instanceof PStatement)
			codegen.appendCommand("pop",-1);
	}
	
	public void inANotEmptyFunctioncall(ANotEmptyFunctioncall node) {
		params.removeAllElements();
		paramStringCall = "";
	}
	
	public void outANotEmptyFunctioncall(ANotEmptyFunctioncall node) {
		Vector paramsfunc = new Vector();
		AId id = (AId) node.getId();
		StringBuffer tempString = new StringBuffer();
		StringBuffer tempString2 = new StringBuffer();
		String name = id.getIdent().getText();
		SymbolTableEntry temp = symtab.getSymbolTableEntry(name);
		for (int i=0; i < temp.getParamIndex(); i++) {
			paramsfunc.addElement(temp.getParamList2(i).getType());
		}
		params.insertElementAt(getTypeAssignment(node.getAssignment()),0);
		if (params.size() < paramsfunc.size()) {
			Minako.error("Function call with too less params");
		} else if (params.size() > paramsfunc.size()) {
			Minako.error("Function call with too many params");
		} else {
			for (int i=0; i < params.size();i++) {
				if (params.get(i) != paramsfunc.get(i))
					Minako.error("Wrong parameter in functioncall at " + (i+1) + ". position");
			}
		}
		for (int i = 0;i < params.size();i++) {
			tempString.append(convertType(params.get(i).toString()));
		}
		paramStringCall = tempString.toString();
		tempString2.append("invokestatic " + Minako.className + "/" + temp.getName() + "(" + paramStringCall + ")" + convertType(temp.getType()));
		codegen.appendCommand(tempString2.toString(),1-params.size());
		if (temp.getType() != Minako.TYP_VOID && (node.parent() instanceof PNoDanglingStatement))
			codegen.appendCommand("pop",-1);
	}
	
	public void outAFuncparams(AFuncparams node) {
		params.addElement(getTypeAssignment(node.getAssignment()));
	}
	
	public void caseAIfelseNoDanglingStatement(AIfelseNoDanglingStatement node) {
        inAIfelseNoDanglingStatement(node);
        if(node.getKwIf() != null)
        {
            node.getKwIf().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getAssignment() != null)
        {
            node.getAssignment().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        symtab.enterBlock();
        if(node.getThen() != null)
        {
            node.getThen().apply(this);
        }
        symtab.leaveBlock();
        if(node.getKwElse() != null)
        {
            node.getKwElse().apply(this);
        }
        symtab.enterBlock();
        if(node.getElse() != null)
        {
            node.getElse().apply(this);
        }
        symtab.leaveBlock();
        if (getTypeAssignment(node.getAssignment()) != Minako.TYP_BOOLEAN)
        	Minako.error("Assignment in if-statement has to be boolean");
        outAIfelseNoDanglingStatement(node);
    }

	public void caseAIfDanglingStatement(AIfDanglingStatement node)
    {
        inAIfDanglingStatement(node);
        if(node.getKwIf() != null)
        {
            node.getKwIf().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getAssignment() != null)
        {
            node.getAssignment().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        symtab.enterBlock();
        if(node.getStatement() != null)
        {
            node.getStatement().apply(this);
        }
        symtab.leaveBlock();
        if (getTypeAssignment(node.getAssignment()) != Minako.TYP_BOOLEAN)
        	Minako.error("Assignment in if-statement has to be boolean");
        outAIfDanglingStatement(node);
    }

	public void caseAIfelseDanglingStatement(AIfelseDanglingStatement node)
    {
        inAIfelseDanglingStatement(node);
        if(node.getKwIf() != null)
        {
            node.getKwIf().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getAssignment() != null)
        {
            node.getAssignment().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        symtab.enterBlock();
        if(node.getNoDanglingStatement() != null)
        {
            node.getNoDanglingStatement().apply(this);
        }
        symtab.leaveBlock();
        if(node.getKwElse() != null)
        {
            node.getKwElse().apply(this);
        }
        symtab.enterBlock();
        if(node.getDanglingStatement() != null)
        {
            node.getDanglingStatement().apply(this);
        }
        symtab.leaveBlock();
        if (getTypeAssignment(node.getAssignment()) != Minako.TYP_BOOLEAN)
        	Minako.error("Assignment in if-statement has to be boolean");
        outAIfelseDanglingStatement(node);
    }

	public void caseAStatassForstatement(AStatassForstatement node)
    {
        inAStatassForstatement(node);
        if(node.getKwFor() != null)
        {
            node.getKwFor().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getStat1() != null)
        {
            node.getStat1().apply(this);
        }
        if(node.getSem1() != null)
        {
            node.getSem1().apply(this);
        }
        if(node.getExpr() != null)
        {
            node.getExpr().apply(this);
        }
        if(node.getSem2() != null)
        {
            node.getSem2().apply(this);
        }
        if(node.getStat2() != null)
        {
            node.getStat2().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        if (getTypeExpr(node.getExpr()) != Minako.TYP_BOOLEAN)
        	Minako.error("Expression in for-statement has to be boolean");
        symtab.enterBlock();
        if(node.getStatement() != null)
        {
            node.getStatement().apply(this);
        }
        symtab.leaveBlock();
        outAStatassForstatement(node);
    }
	
	public void caseADeclassForstatement(ADeclassForstatement node)
    {
        inADeclassForstatement(node);
        if(node.getKwFor() != null)
        {
            node.getKwFor().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getDeclassignment() != null)
        {
            node.getDeclassignment().apply(this);
        }
        if(node.getSem1() != null)
        {
            node.getSem1().apply(this);
        }
        if(node.getExpr() != null)
        {
            node.getExpr().apply(this);
        }
        if(node.getSem2() != null)
        {
            node.getSem2().apply(this);
        }
        if(node.getStatassignment() != null)
        {
            node.getStatassignment().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        if (getTypeExpr(node.getExpr()) != Minako.TYP_BOOLEAN)
        	Minako.error("Expression in for-statement has to be boolean");
        symtab.enterBlock();
        if(node.getStatement() != null)
        {
            node.getStatement().apply(this);
        }
        symtab.leaveBlock();
        outADeclassForstatement(node);
    }
	
	public void caseADowhilestatement(ADowhilestatement node)
    {
        inADowhilestatement(node);
        if(node.getKwDo() != null)
        {
            node.getKwDo().apply(this);
        }
        symtab.enterBlock();
        if(node.getStatement() != null)
        {
            node.getStatement().apply(this);
        }
        symtab.leaveBlock();
        if(node.getKwWhile() != null)
        {
            node.getKwWhile().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getAssignment() != null)
        {
            node.getAssignment().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        if (getTypeAssignment(node.getAssignment()) != Minako.TYP_BOOLEAN)
        	Minako.error("Assignment in dowhile-statement has to be boolean");
        outADowhilestatement(node);
    }
	
	public void caseAWhilestatement(AWhilestatement node)
    {
        inAWhilestatement(node);
        if(node.getKwWhile() != null)
        {
            node.getKwWhile().apply(this);
        }
        if(node.getLparen() != null)
        {
            node.getLparen().apply(this);
        }
        if(node.getAssignment() != null)
        {
            node.getAssignment().apply(this);
        }
        if(node.getRparen() != null)
        {
            node.getRparen().apply(this);
        }
        symtab.enterBlock();
        if(node.getStatement() != null)
        {
            node.getStatement().apply(this);
        }
        if (getTypeAssignment(node.getAssignment()) != Minako.TYP_BOOLEAN)
        	Minako.error("Assignment in while-statement has to be boolean");
        symtab.leaveBlock();
        outAWhilestatement(node);
    }
	
	public void outAIntFactor(AIntFactor node) {
		StringBuffer tempString = new StringBuffer();
		String value = node.getConstInt().getText();
		tempString.append("ldc " + value);
		codegen.appendCommand(tempString.toString(),1);
	}
	
	public void outABooleanFactor(ABooleanFactor node) {
		StringBuffer tempString = new StringBuffer();
		String value = node.getConstBoolean().getText();
		if (value.equals("true")) {
			value = "1";
		} else if (value.equals("false")) {
			value = "0";
		}
		tempString.append("ldc " + value);
		codegen.appendCommand(tempString.toString(),1);
	}
	
	public void outAFloatFactor(AFloatFactor node) {
		StringBuffer tempString = new StringBuffer();
		String value = node.getConstFloat().getText();
		tempString.append("ldc " + value);
		codegen.appendCommand(tempString.toString(),1);
	}
	
	public void outAIdFactor(AIdFactor node) {
		StringBuffer tempString = new StringBuffer();
		AId id = (AId) node.getId();
		SymbolTableEntry current = symtab.getSymbolTableEntry(id.getIdent().getText());
		int type = current.getType();
		if (type == Minako.TYP_INT || type == Minako.TYP_BOOLEAN) {
			if (current.getLevel() == 1) {
				tempString.append("getstatic " + Minako.className + "/" + current.getName() + " " + convertType(type));
			} else {
				tempString.append("iload " + current.getAddress());
			}
		} else if (type == Minako.TYP_FLOAT) {
			if (current.getLevel() == 1) {
				tempString.append("getstatic " + Minako.className + "/" + current.getName() + " F");
			} else {
				tempString.append("fload " + current.getAddress());
			}
		}
		codegen.appendCommand(tempString.toString(),1);
	}
}